(**
   This is a main module of CO3.
@author Naoki Nishida
@since 2014/06/26
@since 2016/08/28
*)

open Ctrs;;

let _ =
  let output = ref "" in
  let result_Un_sound = ref false in
  let result_SRn_sound = ref false in
  let result_U_sound = ref false in
  let result_SR_sound = ref false in
  let append_output s = 
    output := (!output)^"\n"^s 
  in
  try
    (* Check out the number of command-line arguments *)
    if Array.length Sys.argv = 2
    then
      (* Read .trs file *)
      let result = 
	try
	  Readtrs.readtrs (Sys.argv.(1))
	with
	  _ -> raise Unsupported
      in
      (* Check whether an ->= rule is included or non-standard *)
      if has_relrew result || not (is_standard result)
      then raise Unsupported 
      else ();

(* This constraint is for adapting this to CoCo 2016. *)
      (* raise Unsupported if given R is a TRS or not "oriented 1-CTRS" *)
      if is_unconditional result
         || not (is_type_three result && result.ctype = (Some Oriented))
      then raise Unsupported
(* for debug 
      then
	begin
	  match (is_type_one result, result.ctype = (Some Oriented)) with
	    (true,true) ->
	      (print_string "type 1 and oriented\n"; raise Unsupported)
	  | (true,false) ->
	      (print_string "type 1 but not oriented\n"; raise Unsupported)
	  | (false,true) ->
	      (print_string "not type 1 but oriented\n"; raise Unsupported)
	  | (false,false) ->
	      (print_string "neither type 1 nor oriented\n"; raise Unsupported)
	end
*)
      else ();

      (* Decompose conditions: c(s1,...,sn) == c(t1,...,tn) to s1 == t1, ..., sn == tn *)
      let result = decompose_conditions result in
      (* Remove infeasible rules *)
      let result = Infeasibility.rm_infeasible_rules_amap result in
      (* Check first non-confluence *)
      if Cr.is_not_CR result
      then 
	begin
	  append_output "Disproved via the following (C)TRS:";
	  append_output (Ctrs.string_of_ctrs 0 result);
	  print_string "NO\n"
	end
      else (* Case where failed to prove non-confluence of R *)
	(* Check whether R is a TRS or not *)
	if is_unconditional result
	then 
          (* Improved for CoCo 2015 *)
	  (* Check whether R is orthogonal *)
	  if Cr.is_orthogonal result 
	  then 
	    begin
	      append_output "Proved via the following orthogonal TRS:";
	      append_output (Ctrs.string_of_ctrs 0 result);
	      print_string "YES\n"
	    end
	  else
	    (* Check whether R is SN and CP-joinable *)
	    if Sn.is_SN result
	    then
	      if List.for_all (Cpair.is_joinable result) (Cpair.all_cpairs result)
	      then
		begin
		  append_output "Proved via the following terminating and CP-joinable TRS:";
		  append_output (Ctrs.string_of_ctrs 0 result);
		  print_string "YES\n"
		end
	      else 
	    begin
	      append_output "Disproved via the following TRS:";
	      append_output (Ctrs.string_of_ctrs 0 result);
	      print_string "NO\n"
	    end
	    else
	    begin
	      append_output "Failed via the following TRS:";
	      append_output (Ctrs.string_of_ctrs 0 result);
	      print_string "MAYBE\n"
	    end
	else (* Case where R is a CTRS *)
	  begin
	    try
	      (* Check whether SR is sound for R *)
	      if Convert_NormalOneCTRS.is_SR_sound result 
	      then
		result_SRn_sound := true 
	      else 
		();
	      if Convert_NormalOneCTRS.is_U_sound result
	      then 
		result_Un_sound := true
	      else
		();
     	      (* Convert R to SR(R) and U(R) *)
	      let (result2,result3) = Convert_NormalOneCTRS.convert result in
(*
		print_string "SR_n(R) = \n";
		print_string (Ctrs.string_of_ctrs 0 result2);
		print_string "\nU_n(R) = \n";
		print_string (Ctrs.string_of_ctrs 0 result3);
		print_string "\n";
*)
	      if (!result_SRn_sound && Cr.is_CR result2)
	         || (!result_Un_sound && Cr.is_CR result3)
	      then
		print_string "YES\n"
	      else
		raise Not_normal_CTRS
	    with
	      Not_normal_CTRS ->
	        (* Check whether SR is sound for R that is not a normal 1-CTRS  *)
		if Convert_UWLLDCTRS.is_SR_sound result 
		then
		  result_SR_sound := true
		else
		  ();
		if Convert_UWLLDCTRS.is_U_sound result
		then
		  result_U_sound := true
		else
		  ();
	        (* Convert R to SR(R) and U(R) *)
		let (result2,result3) = Convert_UWLLDCTRS.convert result in
(*
		  print_string "SR(R) = \n";
		  print_string (Ctrs.string_of_ctrs 0 result2);
		  print_string "\nU(R) = \n";
		  print_string (Ctrs.string_of_ctrs 0 result3);
		  print_string "\n";
*)
		if (!result_SR_sound && Cr.is_CR result2)
	           || (!result_U_sound && Cr.is_CR result3)
		then
		  print_string "YES\n"
		else 
		  print_string "MAYBE\n"
	  end
    else 
      print_string "usage: co3 <filename>\n"
  with
    Not_3DCTRS ->
      print_string "MAYBE\n" (* for CoCo 2016 *)
      (* print_string "UNSUPPORTED\n" *) (* standard *)
  | Unsupported -> print_string "UNSUPPORTED\n"
(* Commented out for CoCo 
  | Contradiction -> print_string "ERROR: conflicts on CONDITIONTYPE\n"
  | Convert_NormalOneCTRS.UnexpectedError s
  | Convert_UWLLDCTRS.UnexpectedError s -> 
      print_string ("ERROR: unexpected ("^s^")\n")
*)
  | _ -> print_string "MAYBE\n";;
