(**
   This is a module for the Serbanuta & Rosu transformation and the unraveling for normal CTRSs.
   This is an extension of [sr.ml].
   @author Naoki Nishida
   @since 2014/07/03
   @since 2016/08/27
*)
  
open Util;;
open Ctrs;;
  
(**
   This generates reserved symbols.
*)
let reserved f = "#"^f;;
    
let new_x = reserved "x";;
let new_x_max = ref 0;;
let new_z = reserved "z";;
let new_z_max = ref 0;;
let new_v = reserved "v";;
let new_v_max = ref 0;;
let tuple_symbol = reserved "tuple";;
let tuple_max = ref 0;;
let wrap_symbol = (reserved "wrap",None);;
let wrap_max = ref (-1);;
let bottom_symbol = (reserved "bot",None);;
let bottom_max = ref (-1);;
let bracket_symbol = reserved "bra";;
let bracket_max = ref 0;;
let bracket_cntr = ref 0;;
let u_symbol = reserved "u";;
let u_max = ref 0;;
let u_cntr = ref 0;;
let dummy_sym = (reserved "dummy",None);;
    
(**
   This generates a fresh U symbol.
*)
let new_usym () =
  begin
    u_cntr := (!u_cntr) + 1;
    (u_symbol, Some (!u_cntr))
  end;;
    
(**
   This generates a fresh bracket symbol.
*)
let new_bracketsym () =
  begin
    bracket_cntr := (!bracket_cntr) + 1;
    (bracket_symbol, Some (!bracket_cntr))
  end;;
    
let using_opt = ref true;;
let using_reserved = ref true;;
let constructor_based = ref false;;
    
let dsymbols = ref ([] : (fid * int) list);;
let csymbols = ref ([] : (fid * int) list);;
let ex_dsymbols = ref ([] : (fid * (int * int)) list);;
let original_symbols = ref ([] : fid list);;
    
exception UnexpectedError of string;;

(**
   This removes tuple symbols.
*)
let rm_tuple (t1,t2) =
  begin
    match (t1,t2) with
      (Node(Var _,_),Node(Var _,_)) -> ([t1],[t2])
    | (Node(FSym (_,None),_),Node(FSym (_,None),_)) -> ([t1],[t2])
    | (Node(FSym (g1,Some m1),ts1),
       Node(FSym (g2,Some m2),ts2)) ->
	 if g1 = tuple_symbol 
	 then
	   if g2 = tuple_symbol && m1 = m2
	   then (ts1,ts2) else raise (UnexpectedError "")
	 else
	   if g2 = tuple_symbol 
	   then raise (UnexpectedError "")
	   else ([t1],[t2])
    | (Node(FSym (g1,_),_),Node(FSym (g2,_),_)) ->
	if g1 = tuple_symbol || g2 = tuple_symbol
	then raise (UnexpectedError "")
	else ([t1],[t2])
    | (Node(Var _,_),Node(FSym (g2,_),_)) ->
	if g2 = tuple_symbol then raise (UnexpectedError "") else ([t1],[t2])
    | (Node(FSym (g1,_),_),Node(Var _,_)) ->
	if g1 = tuple_symbol then raise (UnexpectedError "") else ([t1],[t2])
  end;;   
    
(**
   This collects all original symbols.
*)
let get_index_max () =
  let rec get_max =
    function
	[] -> ()
      |	((f,None),_)::fs -> get_max fs
      |	((f,Some n),_)::fs ->
	  let g = reserved f in
	  if g = new_x 
	  then
	    if n > !new_x_max 
	    then
	      begin
		new_x_max := n;
		get_max fs
	      end
	    else get_max fs
	  else
	    if g = new_z 
	    then 
	      if n > !new_z_max
	      then
		begin
		  new_z_max := n;
		  get_max fs
		end
	      else get_max fs
	    else 
	      if g = new_v
	      then 
		if n > !new_v_max
		then
		  begin
		    new_v_max := n;
		    get_max fs
		  end
		else get_max fs
	      else 
		if g = tuple_symbol
		then 
		  if n > !tuple_max
		  then
		    begin
		      tuple_max := n;
		      get_max fs
		    end
		  else get_max fs
		else 
    		  if g = fst wrap_symbol
		  then 
		    if n > !wrap_max
		    then
		      begin
			wrap_max := n;
			get_max fs
		      end
		    else get_max fs
		else
		    if g = fst bottom_symbol
		    then 
		      if n > !bottom_max
		      then
			begin
			  bottom_max := n;
			  get_max fs
			end
		      else get_max fs
		    else 
		      if g = u_symbol 
		      then
			if n > !u_max
			then
			  begin
			    u_max := n;
			    get_max fs
			  end
		      else
			  if g = bracket_symbol
			  then
			    if n > !bracket_max
			    then
			      begin
				bracket_max := n;
				get_max fs
			      end
			    else
			      get_max fs
  in
  get_max (!dsymbols);
  get_max (!csymbols);
  ();;


(**
   This collects defined symbols.
*)
let rec get_dsymbols =
  function
      [] -> ()
    | (Rew(Node(FSym f,ts),_,_))::rs ->
	begin
	  try
	    let m = List.assoc f !dsymbols in
	    let n = List.length ts in
	    if n = m then get_dsymbols rs else raise (UnexpectedError "")
	  with
	    Not_found ->
	      begin
		let n = List.length ts in
		dsymbols := (f,n)::(!dsymbols);
		get_dsymbols rs
	      end
	end
    | _ -> raise (UnexpectedError "");;
	  
(**
   This collects constructors from a term.
*)
let rec get_csymbols_in_term =
  function
      Node(Var _,_) -> ()
    | Node(FSym f,ts) ->
	begin
	  try
	    let m = List.assoc f !dsymbols in
	    let n = List.length ts in
	    if m = n
	    then 
	      let _ = List.map get_csymbols_in_term ts in
	      ()
	    else raise (UnexpectedError "")
	  with
	    Not_found ->
	      begin
		try
		  let m = List.assoc f !csymbols in
		  let n = List.length ts in
		  if m = n 
		  then 
		    let _ = List.map get_csymbols_in_term ts in
		    ()
		  else raise (UnexpectedError "")
		with
		  Not_found ->
		    begin
		      csymbols := (f,List.length ts)::(!csymbols);
		      let _ = List.map get_csymbols_in_term ts in
		      ()
		    end
	      end
	end;;
	  
(**
   This collects constructors from conditions.
*)
let rec get_csymbols_in_conditions =
  function
      [] -> ()
    | (Arrow (s,t)::cs) ->
	let _ = get_csymbols_in_term s in
	let _ = get_csymbols_in_term t in
	get_csymbols_in_conditions cs
    | _ -> raise (UnexpectedError "");;
	  
(**
   This collects constructors from rules.
*)
let rec get_csymbols_in_rules =
  function
      [] -> ()
    | (Rew(l,r,c)::rs) ->
	let _ = get_csymbols_in_term l in
	let _ = get_csymbols_in_term r in
	let _ = get_csymbols_in_conditions c in
	get_csymbols_in_rules rs
    | _ -> raise (UnexpectedError "");;
	  
(**
   This decides whether a given term is a constructor term.
*)
let rec is_cterm =
  function
      Node(Var _,_) -> true
    | Node(FSym f,ts) ->
	begin
	  try
	    let _ = List.assoc f (!dsymbols) in
	    false
	  with
	    Not_found -> List.for_all is_cterm ts
	end;;
	  
(**
   This decides whether a given term is basic.
*)
let is_basic =
  function
      Node(Var _,_) -> false
    | Node(FSym f,ts) ->
	begin
	  try
	    let _ = List.assoc f (!dsymbols) in
	    List.for_all is_cterm ts
	  with
	    Not_found -> false
	end;;
	  
(**
   This decides whether rules consist a constructor system.
*)
let rec is_CS =
  function
      [] -> true
    | (Rew(l,_,_)::rs) ->
	is_basic l && is_CS rs
    | _ -> raise (UnexpectedError "");;
	  
(**
   This wraps a term by [wrap_symbol].
*)
let wrap t = if !constructor_based then t else Node(FSym wrap_symbol,[t]);;
    
(**
   This generates auxiliary rules for constructors.
*)
let rec gen_aux_crule (c,m) =
  let rec gen_args n =
    if n > m then [] else (Node(Var (new_x,Some n),[]))::gen_args (n+1)
  in
  let t = Node(FSym c,gen_args 1) in
  let rhs = wrap t in
  let rec mk_subst n =
    let x = (new_x,Some n) in
    if n > m then [] else ([(x,wrap (Node(Var x,[])))])::mk_subst (n+1)
  in
  let thetas = mk_subst 1 in
  List.map (fun theta -> (Rew (apply_subst_to_term theta t,rhs,[]))) thetas
and gen_aux_crules =
  function
      [] -> []
    | (c::cs) ->
	gen_aux_crule c @ gen_aux_crules cs;;
	  
(**
   This puts [bottom_symbol] to extended arguments.
*)
let rec put_bottom t =
  let rec mk_bottomlist i =
    if i <= 0 then [] else (Node(FSym bottom_symbol,[]))::mk_bottomlist (i-1)
  in
  begin
    match t with
      Node(Var _,_) -> t
    | Node(FSym f,ts) ->
	begin
	  try
	    let (m,j) = List.assoc f !ex_dsymbols in
	    let org_args = take m ts in
	    Node(FSym f, (List.map put_bottom org_args)@(mk_bottomlist j))
	  with
	    Not_found ->
	      Node(FSym f, List.map put_bottom ts)
	end
  end;;
    
(**
   This collects f-rules.
*)
let rec get_frules f =
  function 
      [] -> []
    | ((Rew(Node(FSym g,_),_,_) as rl)::rs) ->
	if f = g then rl::(get_frules f rs) else get_frules f rs
    | _ -> raise (UnexpectedError "");;
	  
(**
   Tree-like expression of rules to analyze them.
   l -> r <= s_1 -> t_1, ..., s_k -> t_k
*)
type ruletreeR = 
    Rhs of term (* r *) 
  | CRhs of term (* s_i *) * ruletreeL list
and ruletreeL = 
    CLhs of term (* t_i *) * ruletreeR list
and ruletreeroot = 
    Rule of term (* l *) * ruletreeR list;;

let rec mk_indent n =
  if n <= 0 then "" else "  "^(mk_indent (n-1));;
    
(**
   This shows a given [ruletree].
*)
let rec print_ruletree (Rule (l,rcs)) =
  print_string "Rule Tree:\n";
  print_string ("LHS: "^(string_of_term l));
  print_string " ->\n";
  print_ruletreeR 1 rcs;
  print_string "\n"
and print_ruletreeR n =
  function
      [] -> ()
    | (Rhs r)::rcs -> 
	print_string ((mk_indent n)^"RHS: "^(string_of_term r)^"\n");
	print_ruletreeR n rcs
    | (CRhs (t,scs))::rcs ->
	print_string ((mk_indent n)^"CRHS: "^(string_of_term t)^",\n");
	print_ruletreeL (n+1) scs;
	print_ruletreeR n rcs
and print_ruletreeL n =
  function
      [] -> ()
    | (CLhs (s,tcs))::scs ->
	print_string ((mk_indent n)^"CLHS: "^(string_of_term s)^" ->\n");
	print_ruletreeR (n+1) tcs;
	print_ruletreeL n scs;;

(**
   This analyzes f-rules for SR.
*)
let analyze_for_SR f rs =
  let crules = ref [] in
  let urules = ref [] in
  let rec decompose =
    function
	[] -> 
	  (f,List.rev (!urules),List.rev (!crules))
      |	(Rew(_,_,[]) as rl)::rs ->
	  begin
	    urules := rl::(!urules);
	    decompose rs
	  end
      |	(Rew(l,r,c) as rl)::rs ->
	  begin
	    crules :=  rl::(!crules);
	    decompose rs
	  end
      |	_ -> raise (UnexpectedError "")
  in
  decompose rs;;
    
(**
   This analyzes f-rules for U.
*)
let analyze_for_U f rs =
  let rec mk_ruletreeR r =
    function
	[] -> Rhs r
      |	(Arrow (s,t))::c ->
	  CRhs (s, [CLhs (t, [mk_ruletreeR r c])])
      |	_::_ -> raise (UnexpectedError "")
  in
  let rec insert_rule (l,r,c) =
    function 
	[] -> [Rule (l, [mk_ruletreeR r c])]
      |	((Rule (l1, rc1)) as t)::ts ->
	  if !using_opt && l = l1
	  then
	    (Rule (l1, insert_rc r c rc1))::ts
	  else
	    t::(insert_rule (l,r,c) ts)
  and insert_rc r c0 rcs0 =
    begin
      match rcs0 with
	[] -> [mk_ruletreeR r c0]
      |	((CRhs (s1, rt1)) as sc1)::rcs ->
	  begin
	    match c0 with
	      [] -> (Rhs r)::rcs0
	    | (Arrow (s,t))::c ->
		if !using_opt && s = s1
		then
		  (CRhs (s1, insert_tc (r,c) t rt1))::rcs
		else
		  sc1::(insert_rc r c0 rcs)
	    | _::_ ->
		raise (UnexpectedError "")
	  end
      |	sc1::rcs ->
	  sc1::(insert_rc r c0 rcs)
    end
  and insert_tc (r,c) t tcs0 =
    begin
      match tcs0 with
	[] -> [CLhs(t, [mk_ruletreeR r c])]
      |	((CLhs (t1, sc1)) as tc1)::tcs ->
	  if !using_opt && t = t1
	  then
	    (CLhs (t1, insert_rc r c sc1))::tcs
	  else
	    tc1::(insert_tc (r,c) t tcs)
    end
  in
  let crules = ref [] in
  let urules = ref [] in
  let rec decompose =
    function
	[] -> 
	  (f,List.rev (!urules),List.rev (!crules))
      |	(Rew(_,_,[]) as rl)::rs ->
	  begin
	    urules := rl::(!urules);
	    decompose rs
	  end
      |	(Rew(l,r,c))::rs ->
	  begin
	    crules := insert_rule (l,r,c) (!crules);
	    decompose rs
	  end
      |	_ -> raise (UnexpectedError "")
  in
  decompose rs;;
    
(**
   This collects extended arities of defined symbols.
*)
let rec get_ex_dsymbols =
  function
      [] -> ()
    | (f,_,rs)::fs -> 
	begin
	  try
	    let m = List.assoc f !dsymbols in
	    ex_dsymbols := (f,(m,List.length rs))::(!ex_dsymbols);
	    get_ex_dsymbols fs
	  with
	    Not_found -> raise (UnexpectedError "")
	end;;
	  
let vno = ref 0;;

(**
   These extend terms.
*)
let rec inner_extend =
  function
      (Node(Var _,_) as t) -> t
    |	Node(FSym f,ts) ->
	begin
	  try
	    let (m,j) = List.assoc f (!ex_dsymbols) in
	    let rec gen_ex_args n =
	      if n > j 
	      then [] 
	      else
		begin
		  vno := (!vno) + 1;
		  let num = !vno in
		  Node(Var (new_v,Some num),[])::gen_ex_args (n+1)
		end
	    in
	    Node(FSym f, (List.map inner_extend ts) @ gen_ex_args 1)
	  with
	    Not_found -> 
	      Node(FSym f, List.map inner_extend ts)
	end
and extend t =
  begin
    match t with
      Node(Var _,_) -> t
    | Node(FSym f,ts) ->
	begin
	  try
	    let (m,j) = List.assoc f (!ex_dsymbols) in
	    let rec gen_ex_args n =
	      if n > j then [] else (Node(Var (new_z,Some n),[]))::gen_ex_args (n+1)
	    in
	    Node(FSym f, (List.map inner_extend ts) @ gen_ex_args 1)
	  with
	    Not_found -> 
	      Node(FSym f, List.map inner_extend ts)
	end
  end;;

(**
   This corresponds to the overline operator.
*)
let ol t = put_bottom (extend t);;

(**
   This generates auxiliary rules for defined symbols.
*)
let rec gen_aux_drule (c,(m,j)) =
  let rec gen_args n =
    if n > m then [] else (Node(Var (new_x,Some n),[]))::gen_args (n+1)
  in
  let t = extend (Node(FSym c,gen_args 1)) in
  let rhs = wrap (put_bottom t) in
  let rec mk_subst n =
    let x = (new_x,Some n) in
    if n > m then [] else ([(x,wrap (Node(Var x,[])))])::mk_subst (n+1)
  in
  let thetas = mk_subst 1 in
  List.map (fun theta -> (Rew (apply_subst_to_term theta t,rhs,[]))) thetas
and gen_aux_drules =
  function
      [] -> []
    | (c::cs) ->
	gen_aux_drule c @ gen_aux_drules cs;;
	  
(**
   This generates auxiliary rules.
*)
let gen_aux () =
  if !constructor_based 
  then []
  else
    begin
      let x = Node(Var (new_x,Some 1),[]) in
      (Rew (Node(FSym wrap_symbol,[Node(FSym wrap_symbol,[x])]),(Node(FSym wrap_symbol,[x])),[]))
    end
    ::
    (
    (gen_aux_crules (!csymbols))
    @
    (gen_aux_drules (!ex_dsymbols))
      );;
    
(**
   This renames variables in a given rule in some fixed way.
*)
let rename_rule rl =
  match rl with
    (Rew (l,r,c)) -> 
      begin
	let vno = ref 0 in
	let vrem = ref [] in
	let rec rename_lhs =
	  function
	      Node(Var x,_) ->
		begin
		  try
		    let y = List.assoc x !vrem in
		    Node(Var (new_x,y),[])
		  with
		    Not_found ->
		      begin
			vno := !vno + 1;
			vrem := (x,Some (!vno))::(!vrem);
			Node(Var (new_x, Some (!vno)),[])
		      end
		end
	    | Node(FSym f,ts) ->
		Node(FSym f, List.map rename_lhs ts)
	and rename_rhs =
	  function
	      Node(Var x,_) ->
		begin
		  try
		    let y = List.assoc x !vrem in
		    Node(Var (new_x,y),[])
		  with
		    Not_found -> raise (UnexpectedError "")
		end
	    | Node(FSym f,ts) ->
		Node(FSym f, List.map rename_rhs ts)
	and rename_c =
	  function
	      Arrow(s,t) ->
		Arrow(rename_rhs s,rename_lhs t)
	    | _ -> raise (UnexpectedError "")
	in
	let new_l = rename_lhs l in
	Rew(new_l,rename_rhs r,List.map rename_c c)
      end
  | _ -> raise (UnexpectedError "");;
	
(**
   This converts unconditional rules.
*)
let sru_urule =
  function
      (Rew(l,r,[])) ->
	let lhs = extend l in
	let rhs = wrap (ol r) in
	([Rew (lhs,rhs,[])],[Rew (l,r,[])])
    | _ -> raise (UnexpectedError "");;

(**
   This combines two pairs of rule sets.
*)
let append (xs1,xs2) (ys1,ys2) = (xs1@ys1,xs2@ys2);;

(**
   This converts conditional rules.
*)
let rec sru_crule i f ((l,s),trs) = 
  if trs = [] 
  then ([] ,[])
  else
    let base = extend l in
    let z_i = (new_z,Some i) in
    let bot = Node(FSym bottom_symbol,[]) in
    let lhs = apply_subst_to_term [(z_i, bot)] base in
    let rl1 = Rew (lhs, apply_subst_to_term [(z_i,wrap (ol s))] base, []) in
    let usym = new_usym () in
    let l_vars = 
      List.map (fun x -> Node(Var x,[])) (List.rev (vars_in_term l)) in
    let (s1,_) = rm_tuple (s,s) in
    let rl2 = Rew (l,Node(FSym usym,s1@l_vars),[]) in
    (rl1::
     (List.map 
	(fun (t,r) ->
	  Rew (apply_subst_to_term [(z_i,wrap (ol t))] base, wrap (ol r), []))
	trs),
     rl2::
     (List.map 
	(fun (t,r) ->
	  let (_,t1) = rm_tuple (s,t) in
	  Rew (Node(FSym usym,t1@l_vars), r, []))
	trs))
and sru_crules n f =
  function
      [] -> []
    | (rl::rs) ->
   (sru_crule n f rl)::(sru_crules (n+1) f rs);;

(**
   This converts decomposed rules.
*)
let sru (f,urules,crules) =
  let (rs1,rs2) = 
    List.split ((List.map sru_urule urules)@(sru_crules 1 f (List.rev crules)))
  in
  (List.concat rs1, List.concat rs2);;

(**
   This converts decomposed rules by SR.
*)
let convert_by_SR rs0 = 
  let bot = Node(FSym bottom_symbol,[]) in
  let convert_urule =
    function
	(Rew(l,r,[])) ->
	  let lhs = extend l in
	  let rhs = wrap (ol r) in
	  Rew (lhs,rhs,[])
      | _ -> raise (UnexpectedError "")
  in
  let rec convert_crule i (l0,r0) (t0,vs0,bra) =
    function
	[] -> 
	  let z_i = (new_z,Some i) in
	  let t0 = inner_extend t0 in 
	  let l = 
	    if t0 = bot 
	    then
	      apply_subst_to_term [(z_i, bot)] l0
	    else
	      apply_subst_to_term [(z_i, Node(FSym bra, (wrap t0)::vs0))] l0
	  in
	  [Rew (l,r0,[])]
      |	(Arrow (s,t))::c -> 
	  let z_i = (new_z,Some i) in
	  let xs0 = List.map (fun x -> Node(Var x,[])) (List.rev (vars_in_term t0)) in
	  let xs = union vs0 xs0 in
	  let t0 = extend t0 in (* Need to revise later *)
	  let l = 
	    if t0 = bot 
	    then 
	      apply_subst_to_term [(z_i, bot)] l0
	    else
	      apply_subst_to_term [(z_i, Node(FSym bra, (wrap t0)::vs0))] l0
	  in
	  let s1 = wrap (ol s) in
	  let bsym = new_bracketsym () in
	  let r = 
	    apply_subst_to_term [(z_i, Node(FSym bsym,s1::xs))] l0
	  in
	  (Rew (l,r,[])) :: (convert_crule i (l0,r0) (t,xs,bsym) c)
      |	_::_ -> raise (UnexpectedError "")
  in
  let rec convert_crules n = 
    function
	[] -> []
      |	(Rew (l,r,c))::rs ->
	  (convert_crule n (extend l, wrap (ol r)) (bot,[],dummy_sym) c)
	  @ (convert_crules (n+1) rs)
      |	_::_ ->
	  raise (UnexpectedError "")
  in
  let convert_rule (f,urs,crs) = 
    (List.map convert_urule urs) @ (convert_crules 1 crs)
  in
  List.map convert_rule rs0;;
    
(**
   This converts decomposed rules by U.
*)
let convert_by_U rs0 =
  let rec unravel_rc (Node(f,ts) as l) =
    function
	[] -> []
      |	(Rhs r)::rcs -> (Rew (l,r,[]))::(unravel_rc l rcs)
      |	(CRhs (s, tcs))::rcs ->
	  begin
	    let usym = new_usym () in
	    let l_vars =
	      List.map (fun x -> Node(Var x,[])) (List.rev (vars_in_term l)) in
	    (Rew (l,Node(FSym usym, s :: l_vars),[]))
	    :: (unravel_tc usym l_vars tcs)
	  end
  and unravel_tc f vs =
    function
	[] -> []
      |	(CLhs (t,scs))::tcs ->
	  begin
	    let l = Node(FSym f, t::vs) in
	    (unravel_rc l scs)@(unravel_tc f vs tcs)
	  end
  in 
  let unravel (Rule (l,rcs)) = 
    begin
(*
      print_ruletree (Rule (l,rcs));
*)
      unravel_rc l rcs
    end
  in
  let convert_rule (f,urs,crs) =
    urs @ (List.concat (List.map unravel crs))
  in
  List.map convert_rule rs0;;
      
(**
   This removes reserved symbols from intorduced symbols.
*)
let rec rm_reserved_from_term =
  function
      Node(Var (x,Some n),_) ->
	if x = new_x
	    then Node(Var ("x",Some (n + !new_x_max)),[])
	else if x = new_z
	    then Node(Var ("z",Some (n + !new_z_max)),[])
	else if x = new_v
	    then Node(Var ("v",Some (n + !new_v_max)),[])
	else raise (UnexpectedError "")
    | Node(Var (_,None),_) -> raise (UnexpectedError "")
    | Node(FSym (f,m),ts) ->
	let us = List.map rm_reserved_from_term ts in
	if f = tuple_symbol
	then
	  begin
	    match m with
	      None -> raise (UnexpectedError "")
	    | Some n -> Node(FSym ("tuple",Some (n + !tuple_max)),us)
	  end
        else if f = u_symbol
	then
	  begin
	    match m with
	      None -> raise (UnexpectedError "")
	    | Some n -> Node(FSym ("u",Some (n + !u_max)),us)
	  end
	else if f = bracket_symbol
	then
	  begin
	    match m with
	      None -> raise (UnexpectedError "")
	    | Some n -> Node(FSym ("bra",Some (n + !u_max)),us)
	  end
	else if f = fst wrap_symbol
	then
	  begin
	    match m with
	      None -> 
		if !wrap_max < 0 
		then Node(FSym ("wrap",None),us)
		else Node(FSym ("wrap",Some (!wrap_max + 1)),us)
	    | Some _ -> raise (UnexpectedError "")
	  end
	else if f = fst bottom_symbol
	then
	  begin
	    match m with
	      None -> 
		if !bottom_max < 0 
		then Node(FSym ("bot",None),us)
		else Node(FSym ("bot",Some (!bottom_max + 1)),us)
	    | Some _ -> raise (UnexpectedError "")
	  end
	else Node(FSym (f,m),us)
and rm_reserved_from_rule =
  function
      Rew(l,r,[]) -> 
	Rew(rm_reserved_from_term l, rm_reserved_from_term r,[])
    | _ -> raise (UnexpectedError "")
and rm_reserved rs = 
  if !using_reserved then List.map rm_reserved_from_rule rs else rs;;

(**
    This produces a new TRS.
*)
let new_trs spec rs = mkFromRules (rm_reserved rs);;
(*
   let spec1 = mkFromRules rs in
   { var = union spec.var spec1.var;
   theories = spec.theories;
   rules = rs;
   strategy = spec.strategy;
   others = [];
   ctype = None;
   };;
*)

(**
   This applies SR and the unraveling to a given CTRS.
*)
let convert spec = 
  if is_type_three spec  (* double check *)
  then 
    if is_deterministic spec (* double check *)
    then
      let rs1 = List.map (fun rl -> rename_rule rl) spec.rules in
      begin
	get_dsymbols rs1;
	get_csymbols_in_rules rs1;
	get_index_max ();
(*
	if is_CS rs1 then constructor_based := true else ();
*)
	let (rs2sr,rs2u) =
	  (List.rev 
	    (List.map (fun (f,_) ->
	      analyze_for_SR f (get_frules f rs1)) 
	       (!dsymbols)),
	   List.rev 
	     (List.map (fun (f,_) ->
	       analyze_for_U f (get_frules f rs1)) 
		(!dsymbols))
	     )
	in
	get_ex_dsymbols rs2sr;
	let (sr_rs,u_rs) = 
   (* List.split ((List.map sru rs2) @ [(gen_aux (),[])]) *)
            ((convert_by_SR rs2sr) @ [(gen_aux ())], convert_by_U rs2u)
	in
	(new_trs spec (List.concat sr_rs), new_trs spec (List.concat u_rs))
      end
    else
      raise Not_3DCTRS
  else
    raise Not_3DCTRS;;

(**
   This decides whether the SR transformation is sound for a given CTRS.
*)
let is_SR_sound spec = is_WLL spec && is_UWLL spec;;

(**
   This decides whether the unraveling is sound for a given CTRS.
*)
let is_U_sound spec = is_WLL spec || is_URL spec;;
