(**
   This is a module on dependency pairs arnd related.
@author Naoki Nishida
@since 2014/06/27
*)

open Graph;;
open Ctrs;;

let dsymbols = ref ([] : fid list);;

exception UnexpectedError;;

(**
   This collects defined symbols.
*)
let rec get_dsymbols =
  function
      [] -> ()
    | (Rew(Node(FSym f,ts),_,_))::rs ->
	if List.mem f !dsymbols
	then get_dsymbols rs 
	else
	  begin
	    dsymbols := f::(!dsymbols);
	    get_dsymbols rs
	  end
    | _ -> raise UnexpectedError;;

(**
   This generates the marked symbol.
*)
let marked (f,n) = (f^"#",n);;

(**
    This generates the marked term.
*)
let marked_term =
  function
      Node(Var _,_) -> raise UnexpectedError
    | Node(FSym f,ts) -> Node(FSym (marked f),ts);;

(**
   This collects root defined terms.
*)
let rec get_dterms =
  function
      Node(Var _,_) -> []
    | (Node(FSym f,ts) as t) ->
	(if List.mem f !dsymbols then [t] else [])
	@ (List.concat (List.map get_dterms ts));;

(**
   This computes dependency pairs of a rule.
*)
let dpair =
  function
      Rew(l,r,[]) ->
	let dts = get_dterms r in
	let dts2 = List.filter (fun t -> not (is_proper_subterm t l)) dts in
	let lhs = marked_term l in
	List.map (fun t -> Rew (lhs, marked_term t, [])) dts2
    | _ -> raise Unsupported;;

(**
   This computes the dependency pairs.
*)
let dpairs spec =
  begin
    dsymbols := [];
    get_dsymbols spec.rules;
    List.concat (List.map dpair spec.rules)
  end;;

(**
   This is the CAP known in the DP method.
*)
let cap spec (s,t) =
  begin
    dsymbols := [];
    get_dsymbols spec.rules;
    let xs = vars_in_terms [s;t] in
    begin
      Variable.init_gen_var ("x",xs,[]);
      let rec cap0 t =
	begin
	  match t with
	    Node(Var _,_) -> t
	  | Node(FSym f,ts) ->
	      if List.mem f !dsymbols
	      then Node(Var (Variable.gen_var#generate),[])
	      else Node(FSym f,List.map cap0 ts)
	end
      in
      (cap0 s, cap0 t)
    end
  end;;

let is_edge ds rs (rl1,rl2) =
  begin
    let xs1 = vars_in_rule rl1 in
    Variable.init_gen_var ("x",xs1,[]);
    let rl3 = Rename.rename_rule rl2 in
    begin
      match (rl1,rl3) with
	(Rew(_,Node(f,ts),[]),Rew(Node(g,us),_,[])) -> 
	  f = g
	    &&
	  begin
	    let xs = vars_in_terms (ts@us) in
	    begin
	      Variable.init_gen_var ("x",xs,[]);
	      let rec cap0 t =
		begin
		  match t with
		    Node(Var _,_) -> t
		  | Node(FSym f,ts) ->
		      if List.mem f !dsymbols
		      then Node(Var (Variable.gen_var#generate),[])
		      else Node(FSym f,List.map cap0 ts)
		end
	      and rencap0 t =
		begin
		  match t with
		    Node(Var _,_) -> Node(Var (Variable.gen_var#generate),[])
		  | Node(FSym f,ts) ->
		      if List.mem f !dsymbols
		      then Node(Var (Variable.gen_var#generate),[])
		      else Node(FSym f,List.map cap0 ts)
		end
	      in
	      let (ts1,us1) = (List.map rencap0 ts, us) in
	      begin
		try
		  let _ = unify (Node(f,ts1)) (Node(g,us1)) in
		  true
		with
		  Not_found -> false
	      end
	    end
	  end
      | _ -> raise Unsupported
    end
  end;;	

let rec print_nodes =
  function
      [] -> ()
    | (n,rl)::ns ->
	begin
	  print_string ("    ("^(string_of_int n)^") ");
	  print_string (string_of_rule None rl);
	  print_string "\n";
	  print_nodes ns
	end;;

let rec print_edges =
  function
      [] -> ()
    | ((n1,n2)::es) ->
	begin
	  print_string
	    ("    ("^(string_of_int n1)^") -> ("^(string_of_int n2)^")\n");
	  print_edges es
	end;;

let print_edg g =
  begin
    print_string "(COMMENT EDG is\n";
    print_string "  NODES:\n";
    print_nodes g.nodes;
    print_string "\n";
    print_string "  EDGES:\n";
    print_edges g.edges;
    print_string ")\n"
  end;;

let mk_edg spec dps =
  begin
    dsymbols := [];
    get_dsymbols spec.rules;
    let g = mk_graph dps (is_edge (!dsymbols) spec.rules) in
    begin
(*
   print_edg g;
*)
      g
    end
  end;;
